/* VS1033D Decoder Demo
 * http://frank.circleofcurrent.com/musicclock/
 * Copyright (c) 2011 Frank Zhao
 * 
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of either the GNU General Public License version 2
 * or the GNU Lesser General Public License version 2.1, both as
 * published by the Free Software Foundation.
 *
 */

// please see http://www.nongnu.org/avr-libc/user-manual/modules.html for info on these modules

#include <avr/io.h> // allows access to AVR built-in peripherals and registers
#include <avr/pgmspace.h> // allows access to flash memory
#include <util/delay.h> // provides delays
#include <stdio.h> // allows formatted debug output
#include <avr/power.h> // used for setting clock prescaler
#include "libs/ff.h" // use FatFs from Elm-Chan.org
#include "libs/diskio.h" // use FatFs from Elm-Chan.org (diskio used for disk initialization)

#define USB_DEBUG_ENABLE // use this to switch the debug on or off

#ifdef USB_DEBUG_ENABLE
#include "libs/usb_serial.h" // provides USB-to-serial functionality
#endif

// pin definitions
#define MP3_PORTDIR DDRC
#define MP3_PORTOUTPUT PORTC
#define MP3_PORTINPUT PINC
#define MP3_PIN_CS 4
#define MP3_PIN_DCS 5
#define MP3_PIN_RST 7
#define MP3_PIN_DREQ 6

// macros to change SPI speed
#define spi_slowest() SPCR |= _BV(SPR1) | _BV(SPR0); SPSR &= ~_BV(SPI2X)
#define spi_slower() SPCR = (SPCR & (~_BV(SPR1))) | _BV(SPR0); SPSR |= _BV(SPI2X)
#define spi_slow() SPCR &= ~_BV(SPR1) & ~_BV(SPR0); SPSR &= ~_BV(SPI2X)
#define spi_fast() SPCR &= ~_BV(SPR1) & ~_BV(SPR0); SPSR |= _BV(SPI2X)
// spi_slow is the fastest that we can operate the VS1033D at
// spi_fast is the fastest that we can operate the MicroSD card at

// this is a wrapper for usb_putchar that allows me to use printf from stdio.h
static inline int debug_putchar(char c, FILE* stream)
{
    #ifdef USB_DEBUG_ENABLE
    return usb_serial_putchar(c); // refer to http://www.pjrc.com/teensy/usb_serial.html
    #else
    return 0;
    #endif
}

// setup stream for printf from stdio.h
static FILE mystdout = FDEV_SETUP_STREAM(debug_putchar, NULL, _FDEV_SETUP_WRITE);

typedef enum
{
    MP3_REG_MODE = 0x00,
    MP3_REG_CLOCKF = 0x03,
    MP3_REG_VOL = 0x0B,
}
mp3_register_t;

typedef enum
{
    PLAYSTATUS_OK,
    PLAYSTATUS_BUSY,
    PLAYSTATUS_ENDOFFILE,
}
mp3_playStatus_t;

inline void mp3_waitReady(uint8_t cnt)
{
    for(uint8_t i = 0; i < cnt; i++)
	{
		_delay_us(10);
		if (bit_is_clear(MP3_PORTINPUT, MP3_PIN_DREQ)) break;
	}
	loop_until_bit_is_set(MP3_PORTINPUT, MP3_PIN_DREQ);
}

inline uint8_t mp3_spi(uint8_t data)
{
    SPDR = data; // send
    loop_until_bit_is_set(SPSR, SPIF); // wait until sent
    return SPDR; // get reply
}

void mp3_writeReg(mp3_register_t addr, uint16_t data)
{
    spi_slow();
    
    MP3_PORTOUTPUT &= ~_BV(MP3_PIN_CS); // select
    
    mp3_spi(0b00000010); // write command
    mp3_spi(addr); // send address
    mp3_spi((data & 0xFF00) >> 8); // send data MSB first
    mp3_spi(data & 0x00FF);
    
    MP3_PORTOUTPUT |= _BV(MP3_PIN_CS); // deselect
    
    spi_fast();
    
    mp3_waitReady(10);
}

void mp3_playPacket(uint8_t* buffer, uint8_t length)
{
    spi_slow();
    loop_until_bit_is_set(MP3_PORTINPUT, MP3_PIN_DREQ); // wait for ready
    MP3_PORTOUTPUT &= ~_BV(MP3_PIN_DCS); // select
    for (uint8_t i = 0; i < length; i++)
    {
        mp3_spi(buffer[i]);
    }
    MP3_PORTOUTPUT |= _BV(MP3_PIN_DCS); // deselect
    spi_fast();
}

mp3_playStatus_t mp3_playFile(FIL* file)
{
    uint8_t buff[32];
    UINT read;
    f_read(file, buff, 32, &read); // read the file
    
    mp3_playPacket(buff, read); // play the sound
    
    if (read != 32) return PLAYSTATUS_ENDOFFILE;
    
    if (MP3_PORTINPUT & _BV(MP3_PIN_DREQ) == 0) return PLAYSTATUS_BUSY;
    
    return PLAYSTATUS_OK;
}

void mp3_init()
{
    // note that this must be called after the SD card has initialized to ensure SPI has initialized

    // setup all pin directions
    
    // reset pin, hold it low during setup
    MP3_PORTDIR |= _BV(MP3_PIN_RST);
    MP3_PORTOUTPUT &= ~_BV(MP3_PIN_RST);
    
    // DREQ is input, no pull-up
    MP3_PORTDIR &= ~_BV(MP3_PIN_DREQ);
    MP3_PORTOUTPUT &= ~_BV(MP3_PIN_DREQ);
    
    // both chip selects as output, starting high
    MP3_PORTDIR |= _BV(MP3_PIN_CS) | _BV(MP3_PIN_DCS);
    MP3_PORTOUTPUT |= _BV(MP3_PIN_CS) | _BV(MP3_PIN_DCS);
    
    _delay_us(10); // slight delay
    MP3_PORTOUTPUT |= _BV(MP3_PIN_RST); // bring out of reset, get ready
    
    mp3_waitReady(10);    
	
	mp3_writeReg(MP3_REG_MODE, 0b0000100000000011); // see datasheet
	mp3_writeReg(MP3_REG_CLOCKF, 0x1800 | 0xE000); // see datasheet
}

// program entry point
int main()
{
    clock_prescale_set(clock_div_2); // this is because we are running at 3.3V
    // see http://www.nongnu.org/avr-libc/user-manual/group__avr__power.html and http://www.pjrc.com/teensy/3volt.html
    
    stdout = &mystdout; // setup stream for printf from stdio.h
    
    // initialize stuff
    #ifdef USB_DEBUG_ENABLE
    usb_init();
    
    while (usb_configured() == 0) _delay_ms(5); // waits until USB is ready
    
    _delay_ms(5000); // 5 seconds for you to connect hyperterminal
    #endif
    
    // here's the start of the demonstration
    printf_P(PSTR("VS1033D Decoder Demo\r\n"));
    
    FATFS filesystem;
    DIR rootDir;
    FRESULT errCode = -1;
    
    while (errCode != FR_OK)
    {
        printf_P(PSTR("Mounting File System\r\n"));
        errCode = f_mount(0, &filesystem);
        if (errCode != FR_OK)
        {
            printf_P(PSTR("Error: f_mount failed, error code: %d\r\n"), errCode);
            _delay_ms(1000); // retry after 1 sec
            continue;
        }
        
        printf_P(PSTR("Opening Root Directory\r\n"));
        errCode = f_opendir(&rootDir, "/");
        if (errCode != FR_OK)
        {
            printf_P(PSTR("Error: f_opendir failed, error code: %d\r\n"), errCode);
            _delay_ms(1000); // retry after 1 sec
            continue;
        }
    }
    
    printf_P(PSTR("Initializing VS1033D\r\n"));
    mp3_init();
    mp3_writeReg(MP3_REG_VOL, 0); // max volume, attenuation = 0 means max volume, see datasheet
    
    // read all file entries
    while (1)
    {
        FILINFO fileEntry;
        errCode = f_readdir(&rootDir, &fileEntry);
        if (errCode != FR_OK)
        {
            printf_P(PSTR("Error: f_readdir failed, error code: %d\r\n"), errCode);
            while (1); // freeze
        }
        
        if (fileEntry.fname[0] == 0)
        {
            printf_P(PSTR("End of All Files\r\n"));
            break;
        }
        else if (fileEntry.fname[0] == '.')
        {
            printf_P(PSTR("Skipping \"%s\"\r\n\r\n"), fileEntry.fname);
        }
        else if (fileEntry.fattrib & AM_DIR)
        {
            printf_P(PSTR("Directory Entry: %s\r\n\r\n"), fileEntry.fname);
        }
        else
        {
            printf_P(PSTR("File Entry: %s\r\n"), fileEntry.fname);
            
            // open file
            FIL file;
            errCode = f_open(&file, fileEntry.fname, FA_READ);
            if (errCode != FR_OK)
            {
                printf_P(PSTR("Error: f_open failed, error code: %d\r\n"), errCode);
                while (1); // freeze
            }
            
            // read file until end
            while (mp3_playFile(&file) != PLAYSTATUS_ENDOFFILE);
            
            printf_P(PSTR("\r\nEnd of File\r\n\r\n"));
            
            errCode = f_close(&file);
            if (errCode != FR_OK)
            {
                printf_P(PSTR("Error: f_close failed, error code: %d\r\n"), errCode);
            }
        }
    }
    
    while (1); // infinite loop
    return 0;
}